/**
 * Copyright (c) 2004 Logos GmbH Rostock (www.logos-ib.de).
 * All rights reserved.
 *
 * THIS SOFTWARE IS  PROVIDED  "AS IS"  AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF  MERCHANTABILITY  AND  FITNESS  FOR  A PARTICULAR  PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL  LOGOS GMBH  OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES  (INCL., BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR  PROFITS;
 * OR BUSINESS INTERRUPTION)  HOWEVER  CAUSED  AND  ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,  OR TORT (INCL.
 * NEGLIGENCE OR OTHERWISE)  ARISING  IN  ANY  WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



package de.inovat.kex.bast.exporter;

//~ NICHT JDK IMPORTE =========================================================

import de.bsvrz.dav.daf.main.Dataset;
import de.inovat.kex.bast.SendeBAStStatusDaten;
import de.inovat.kex.bast.common.ChannelException;
import de.inovat.kex.bast.common.objectwrapper.LangZeitZaehlStelle;

import java.text.DecimalFormat;

//~ JDK IMPORTE ===============================================================

//~ KLASSEN ===================================================================

/**
 * stellt die Funktionalit&auml;t bereit die von allen Channels zur Ausgabe von
 * BASt-Formaten ben&ouml;tigt werden.
 *
 * <p>Vor der Bearbeitung der eigentlichen Daten durch die Aufrufe der Methode
 * { @link ResultDataChannel#processData()}, muss
 * der Channel vollst&auml;ndig initialisiert sein.</p>
 *
 * @author LOGOS GmbH Rostock (www.logos-ib.de)
 * @author R. Trier
 * @version $Id: AbstractDataChannelBASt.java 452 2008-11-10 07:09:39Z Schawaller $
 */
abstract public class AbstractDataChannelBASt extends FileDataChannel {

    /**
     * Array f&uuml;r MonatsCodes 1=Jan .. c=Dez (laut Tabelle der BASt)
     */
    protected static final char[] MONATE = new char[] { '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c' };

    /**
     * DecimalFormat initialisiert mit "00"
     */
    public static DecimalFormat TWO_DIGIT_FORMAT = new DecimalFormat("00");

    //~ FELDER ================================================================

    /**
     * die letzte bearbeitete Stunde (um Sommer-/Winterzeitumstellung zu
     * behandeln).
     */
    protected int _LastHour = -1;

    /**
     * Anzahl der Fahrstreifen der aktuellen Z&auml;hlstelle, wird in
     * setZaehlstelle berechnet.
     */
    protected int _CountOfFS;

    /**
     * aktuelle Langzeitz&auml;hlstelle
     */
    protected LangZeitZaehlStelle _ZaehlStelle;

    //~ KONSTRUKTOREN  (und vom Konstruktor verwendete Methoden) ==============

    /**
     * erzeugt einen DataChannel, der mit einer reset-Methode und eventuell {@link #setZaehlstelle(LangZeitZaehlStelle zaehlstelle) setZaehlstelle()}
     * initialisiert werden muss.
     *
     * @param bastStatusDaten Statusinformationen.
     */
    public AbstractDataChannelBASt(SendeBAStStatusDaten bastStatusDaten) {
        super(bastStatusDaten);
    }

    //~ METHODEN ==============================================================

    /**
     * gibt den Dateinamen zur&uuml;ck, der sich
     * aus den Daten der Z&auml;hlstelle, dem Jahr, dem Monat ergibt.
     *
     * @param zaehlStelle Zählstelle.
     * @param year Jahr.
     * @param month  Monat.
     *
     * @return vollst&auml;ndig klassifizierter Dateiname
     */
    abstract protected String constructFilename(LangZeitZaehlStelle zaehlStelle, int year, int month);

    /**
     * gibt den Dateinamen zur&uuml;ck, der sich
     * aus den Daten der Z&auml;hlstelle, dem Jahr, dem Monat und dem Tag ergibt.
     *
     * @param zaehlStelle Zählstelle.
     * @param year  Jahr.
     * @param month Monat.
     * @param day  Tag.
     *
     * @return  Filename als String.
     */
    abstract protected String constructFilename(LangZeitZaehlStelle zaehlStelle, int year, int month, int day);

    //~ GET METHODEN ==========================================================

    /**
     * <p>Hilfsmethode, die pr&uuml;ft, ob das ResulData[] irgendwelche Data
     * hat.</p>
     * <p>Funktion wird verwendet, um zu pr&uuml;fen, ob ein Totalausfall einer Z&auml;hlstelle vorliegt.
     *
     * @param rsArray zu pr&uuml;fendes ResultData[]
     * @return true, wenn das Array hasData, sonst false.
     */
    public static boolean hasData(Dataset[] rsArray) {
        boolean retValue = false;

        if ((rsArray != null) && (rsArray.length > 0)) {
            for (int i = 0; ((i < rsArray.length) && (!retValue)); i++) {
                retValue = ((rsArray[i] != null) && (rsArray[i].getData() != null));
            }
        }

        return retValue;
    }

    //~ METHODEN ==============================================================

    /**
     * <p>ist der Channel ge&ouml;ffnet, werden gepufferte Daten in die Datei
     * geschrieben und der Channel zur Ausgabe in die Datei neu ge&ouml;ffnet,
     * dessen Namen durch die Implementation der Methode
     * { @link #constructFilename(LangZeitZaehlStelle zaehlStelle, String path, int year, int monthOrQuartal, boolean isQuartal) constructFilename}
     * aus den Daten der Z&auml;hlstelle, Jahr, Monat bzw Quartal bestimmt wird.</p>
     *
     * @param zaehlStelle LangZeitZaehlStelle dessen Daten ausgegeben werden
     *   sollen.
     * @param year das Jahr für das die Datei erzeugt werden soll.
     * @param month der Monat für das die Datei erzeugt werden soll (Januar=0).
     * @throws ChannelException .
     */
    public void reset(LangZeitZaehlStelle zaehlStelle, int year, int month) throws ChannelException {
        String unterordner = String.format("%02d%02d", year - 2000, month + 1);
        String fileName    = constructFilename(zaehlStelle, year, month);

        reset(zaehlStelle, fileName, unterordner);
    }

    /**
     * Ist der Channel ge&ouml;ffnet, werden gepufferte Daten in die Datei
     * geschrieben und der Channel zur Ausgabe in die Datei neu ge&ouml;ffnet.
     *
     * @param zaehlStelle LangZeitZaehlStelle deren Daten ausgegeben werden sollen.
     * @param fileName Dateiname.
     * @param unterOrdner Unterverzeichnis, wo die Datei abgelegt werden (siehe {@link #_unterordnerName}).
     *
     * @throws ChannelException Exception.
     */
    private void reset(LangZeitZaehlStelle zaehlStelle, String fileName, String unterOrdner) throws ChannelException {
        setZaehlstelle(zaehlStelle);
        reset(fileName, unterOrdner);
    }

    /**
     * <p>ist der Channel ge&ouml;ffnet, werden gepufferte Daten in die Datei
     * geschrieben und der Channel zur Ausgabe in die Datei neu ge&ouml;ffnet,
     * dessen Namen durch die Implementaiton der Methode
     * { @link #constructFilename(LangZeitZaehlStelle zaehlStelle, String path, int year, int startDayOfYear, int endDayOfYear) constructFilename}
     * aus den Daten der Z&auml;hlstelle, Jahr, Monat bzw Quartal bestimmt wird.</p>
     *
     * @param zaehlStelle LangZeitZaehlStelle dessen Daten ausgegeben werden
     *   sollen.
     * @param year das Jahr für das die Datei erzeugt werden soll.
     * @param month der Monat für das die Datei erzeugt werden soll.
     * @param day der erste Tag des Monats, f&uuml;r den die Daten ausgegeben werden.
     * @throws ChannelException .
     */
    public void reset(LangZeitZaehlStelle zaehlStelle, int year, int month, int day) throws ChannelException {
        String unterordner = String.format("%02d%02d", year - 2000, month);
        String fileName    = constructFilename(zaehlStelle, year, month, day);

        reset(zaehlStelle, fileName, unterordner);
    }

    //~ SET METHODEN ==========================================================

    /**
     * <p>setzt die Z&auml;hlstelle.</p>
     *
     * @param zaehlstelle LangZeitZaehlStelle
     * @throws ChannelException .
     */
    abstract public void setZaehlstelle(LangZeitZaehlStelle zaehlstelle) throws ChannelException;
}


//~Formatiert mit 'inovat Kodierkonvention' am 06.08.12
