/*
 * Copyright (c) 2010 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.inovat.kex.bast.exporter.BAStExportSteuerung
 *
 * de.inovat.kex.bast.exporter.BAStExportSteuerung is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * de.inovat.kex.bast.exporter.BAStExportSteuerung is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * <p>
 * You should have received a copy of the GNU General Public License
 * along with de.inovat.kex.bast.exporter.BAStExportSteuerung; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * Koelner Strasse 30
 * D-50859 Köln
 * +49 (0)2234 4301 800
 * info@invat.de
 * www.inovat.de
 */

package de.inovat.kex.bast.exporter;

//~ NICHT JDK IMPORTE =========================================================

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;

import de.inovat.kex.bast.common.DataModellMismatchException;
import de.inovat.kex.bast.common.NetzClientDavConnection;
import de.inovat.kex.bast.common.objectwrapper.LangZeitZaehlStelle;
import de.inovat.kex.bast.KonstantenBast.BAST_FORMAT;
import de.inovat.kex.bast.SendeBAStStatusDaten;

//~ JDK IMPORTE ===============================================================

import java.text.ParseException;
import java.text.SimpleDateFormat;

import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

//~ KLASSEN ===================================================================

/**
 * Die Klasse dient als Zwischenklasse zur Steuerung des Exportes durch den Datenverteiler.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dagmar Rostek (DR)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class BAStExportSteuerung implements StandardApplication, ClientReceiverInterface, Runnable {
    private static final String ABBRUCH = "Der Export wurde abgebrochen.  ";

    /** Aspekt Anfrage. */
    private static final String ASP_ANFRAGE = "asp.anfrage";

    /** Attributgruppe zur Steuerung des BASt-Band Exports. */
    private static final String ATG_EXPORT_STEUERUNG                                                      = "atg.bastExportSteuerung";
    public static final String  ATG_tlsAxlEinzelErgebnisMeldungAchsLastenUndKfzGesamtGewichteVersion1Bis2 = "atg.tlsAxlEinzelErgebnisMeldungAchsLastenUndKfzGesamtGewichteVersion1Bis2";
    public static final String  ATG_tlsAxlEinzelErgebnisMeldungAchsLastenUndKfzGesamtGewichteVersion3     = "atg.tlsAxlEinzelErgebnisMeldungAchsLastenUndKfzGesamtGewichteVersion3";

    /** Debugdebug fuer Debug-Ausgaben. */
    private static Debug                  debug           = Debug.getLogger();
    private static final List<ResultData> _listeAuftraege = new ArrayList<>();
    private static int                    _versionAxlEinzeldaten;

    /**
     * Alternative zur DaV-Konfiguration KlassifizierungsgruppeBAST oder null,
     * falls die Daten aus DaV-Konfiguration gelesen werden soll.
     */
    private static String _parKlassifizierungsgruppeBAST;

    /**
     * Parameter definiert, wie der DateiName für die Version2004 gebildet wird:
     * true - aus der LZZS-Nummer,
     * false - aus der Bast-Nummer.
     */
    private static boolean _isDateiNameV2004LzzsNr;

    //~ FELDER ================================================================

    /**
     * Boolesches Attribut das angibt, ob der Export unterbrochen wurde, so dass keine weitere Statusmeldung mehr gesendet werden darf.
     */
    private boolean _abbruchExport = false;

    /**
     * Endedatum der zu exportierenden Langzeitzählstellen.
     */
    private Calendar _CalEnd = new GregorianCalendar();

    /**
     * Startdatum der zu exportierenden Langzeitzählstellen.
     */
    private Calendar _CalStart = new GregorianCalendar();

    /**
     * Boolesches Attribut das angibt ob der Threat bzgl. BASt-Export gerade aktiv ist.
     */
    private boolean _threadAktiv = false;

    /**
     * Instanz BastExporter.
     */
    private BastExporter _BastExporter;

    /**
     * Exportformat.
     */
    private BAST_FORMAT _bastFormat;

    /** Statusinformationen die an den Datenverteiler gesendet werden. */
    private SendeBAStStatusDaten _bastStatusDaten;

    /**
     * Datenbeschreibungsobjekt Export Steuerung.
     */
    private DataDescription _datenbeschreibungExportSteuerung;

    /** Datenverteilerverbindung */
    private ClientDavInterface _dav;

    /**
     * Host zum alternativen DaV für die Archiv-Daten oder null,
     * falls aktuelle DaV für Archiv-Anfragen verwendet wird.
     * <p>
     * ACHTUNG: Spezieller Anwendungsfall für BW, um die BASt-Daten auf Basis
     * der Archivdaten der VRZ zu erzeugen (davArchivDatenverteiler), der Prozess aber
     * weiterhin auf der DAUZ gestartet werden kann (da hier die Konfig für die Steuerung
     * vorhanden ist und die Bedienung auf diese zugreift).
     * Sonderfall wurde notwendig, da BW-DAUZ Archiv plötzlich gelöscht war, aber die BASt-Daten
     * für das Jahr 2015 und 2016 nachgerechnet werden musste.
     */
    private String _davArchivDatenverteiler;

    /** Benutzer zum alternativen DaV für die Archiv-Daten (s.{@link #_davArchivDatenverteiler}). */
    private String _davArchivBenutzer;

    /** Passwort zum alternativen DaV für die Archiv-Daten (s.{@link #_davArchivDatenverteiler}). */
    private String _davArchivPasswort;

    /**
     * Liste der Pidnummern der zu exportierenden Langzeitzählstellen.
     */
    private List<String> _LzzPidList;

    /** Integer Variable die dem Monat bis entspricht. */
    private int _monatB;

    /** Integer Variable die dem Monat von entspricht. */
    private int _monatV;

    /**
     * Systemobjekt für den BASt-Band Export.
     */
    private SystemObject _systemObjectBAStExportImportSteuerung;

    /** Bis Tag. */
    private int _tagBis;

    /** Vom Tag. */
    private int _tagVon;

    /**
     * Thread fü rden BastBand  Export.
     */
    private Thread _threadBastBandExport;

    /** Typ des Systemobjekte für den BASt-Band Import/Export. */
    private String _typBAStBand;

    //~ METHODEN ==============================================================

    /**
     * Anmeldung beim Datenverteiler als Empfänger für die Attribute bzgl. BASt-Band Export.
     */
    private void anmeldenDaVEmpfaengerBAStExport() {
        AttributeGroup attributgruppeBAStExport;
        Aspect         aspektZustand;
        ReceiveOptions anmeldeoptionen = ReceiveOptions.normal();
        ReceiverRole   empfaengerrolle = ReceiverRole.drain();

        _systemObjectBAStExportImportSteuerung = _dav.getDataModel().getObject(_typBAStBand);
        attributgruppeBAStExport               = _dav.getDataModel().getAttributeGroup(ATG_EXPORT_STEUERUNG);
        aspektZustand                          = _dav.getDataModel().getAspect(ASP_ANFRAGE);
        _datenbeschreibungExportSteuerung      = new DataDescription(attributgruppeBAStExport, aspektZustand);

        // Anmelden zum Empfang der Daten
        subscribe(_systemObjectBAStExportImportSteuerung, _datenbeschreibungExportSteuerung, anmeldeoptionen, empfaengerrolle);
    }

    /**
     * Arbeite Aufträge ab.
     */
    private void arbeiteAuftraegeAb() {
        while (true) {
            if (!_threadAktiv) {
                synchronized (_listeAuftraege) {
                    if (_listeAuftraege.size() > 0) {
                        ResultData auftrag = _listeAuftraege.get(0);

                        _listeAuftraege.remove(0);

                        try {
                            konvertiereDaVDatenInInternesFormat(auftrag);
                        }
                        catch (DataModellMismatchException e) {
                            debug.error(String.format("Es sind Probleme bei der Konvertierung der Daten in das interne Format aufgetreten.%s", e));
                        }
                    }
                }
            }

            try {
                Thread.sleep(10000);  // Alle 10 Sekunden reicht
            }
            catch (InterruptedException e) {

                // nichts tun
            }
        }
    }

    /**
     * Erstelle die Kalender-Instanz aus dem Muster-Kalender mit dem Monatsanfang.
     *
     * @param calenderMuster Muster-Kalender.
     *
     * @return Kalender-Instanz aus dem Muster-Kalender mit dem Monatsanfang.
     */
    public static Calendar erstelleKalenderMitMonatAnfang(Calendar calenderMuster) {
        Calendar calender = new GregorianCalendar();

        calender.setTimeInMillis(calenderMuster.getTimeInMillis());

        // Tag
        calender.set(Calendar.DATE, 1);

        // Zeit:
        calender.set(Calendar.HOUR_OF_DAY, 0);
        calender.set(Calendar.MINUTE, 0);
        calender.set(Calendar.SECOND, 0);
        calender.set(Calendar.MILLISECOND, 0);

        return calender;
    }

    /**
     * Erstelle die Kalender-Instanz aus dem Muster-Kalender mit dem Kalenderende.
     *
     * @param calenderMuster Muster-Kalender.
     *
     * @return Kalender-Instanz aus dem Muster-Kalender mit dem Kalenderende.
     */
    public static Calendar erstelleKalenderMitMonatEnde(Calendar calenderMuster) {
        Calendar calender = new GregorianCalendar();

        calender.setTimeInMillis(calenderMuster.getTimeInMillis());

        // Tag
        calender.set(Calendar.DATE, calenderMuster.getActualMaximum(Calendar.DAY_OF_MONTH));

        // Zeit
        calender.set(Calendar.HOUR_OF_DAY, 23);
        calender.set(Calendar.MINUTE, 59);
        calender.set(Calendar.SECOND, 0);
        calender.set(Calendar.MILLISECOND, 0);

        return calender;
    }

    /**
     * Erstelle die Kalender-Instanz aus dem Muster-Kalender mit der Zeit 00:00:00:00.
     *
     * @param calenderMuster Muster-Kalender.
     *
     * @return Kalender-Instanz aus dem Muster-Kalender mit der Zeit 00:00:00:00.
     */
    public static Calendar erstelleKalenderMitZeitAnfang(Calendar calenderMuster) {
        Calendar calender = new GregorianCalendar();

        calender.setTimeInMillis(calenderMuster.getTimeInMillis());
        calender.set(Calendar.HOUR_OF_DAY, 0);
        calender.set(Calendar.MINUTE, 0);
        calender.set(Calendar.SECOND, 0);
        calender.set(Calendar.MILLISECOND, 0);

        return calender;
    }

    /**
     * Erstelle die Kalender-Instanz aus dem Muster-Kalender mit der Zeit 23:59:59:999.
     *
     * @param calenderMuster Muster-Kalender.
     *
     * @return Kalender-Instanz aus dem Muster-Kalender mit der Zeit 23:59:59:999.
     */
    public static Calendar erstelleKalenderMitZeitEnde(Calendar calenderMuster) {
        Calendar calender = new GregorianCalendar();

        calender.setTimeInMillis(calenderMuster.getTimeInMillis());
        calender.set(Calendar.HOUR_OF_DAY, 23);
        calender.set(Calendar.MINUTE, 59);
        calender.set(Calendar.SECOND, 59);
        calender.set(Calendar.MILLISECOND, 999);

        return calender;
    }

    private void exportFTPAxlEF1() throws SendSubscriptionNotConfirmed {
        debug.info("Der Export Vorgang der Achslastdaten wird gestartet.");
        FTPServerConnection.getInstance().initialisiereProgressbarKonstanten();

        if (_LzzPidList != null) {
            for (String pid : _LzzPidList) {
                if (!_abbruchExport) {
                    LangZeitZaehlStelle lzz = null;

                    try {
                        lzz = LangZeitZaehlStelle.getInstance(pid);
                    }
                    catch (Exception e) {
                        sendeMeldungFehler("", String.format("Die Achslastmessstelle mit der Pid %s Konnte nicht ermittelt werden!", pid));
                    }

                    if (lzz == null) {
                        sendeMeldungFehler("", String.format("Es existieren keine Achslastmessstelle mit der Pid %s. Der Export für den Datensatz findet nicht statt!\"", pid));
                    } else if (isAchslastMessstelle(lzz)) {
                        if (!_abbruchExport) {
                            sendeMeldungInfo("", String.format("Die Daten der Achslastmessstelle mit der Pid: %s werden exportiert.", lzz));

                            // -----------------------------------------------------------------------
                            // Definiere End-Zeitpunkt:
                            Calendar calBis = erstelleKalenderMitZeitAnfang(_CalEnd);

                            calBis.set(Calendar.DATE, _tagBis);

                            // -----------------------------------------------------------------------
                            // Iteriere durch alle Tage:
                            Calendar calTagAnfang = erstelleKalenderMitZeitAnfang(_CalStart);

                            calTagAnfang.set(Calendar.DATE, _tagVon);

                            Calendar calTagEnde;

                            while (calTagAnfang.getTimeInMillis() <= calBis.getTimeInMillis()) {

                                //
                                if (!_abbruchExport) {
                                    calTagEnde = erstelleKalenderMitZeitEnde(calTagAnfang);

                                    try {
                                        _BastExporter.export(lzz, calTagAnfang, calTagEnde);
                                    }
                                    catch (Exception e) {
                                        String infoText = String.format("Es sind Probleme beim Export aufgetreten. Der aktuelle Exportvorgang wird abgebrochen. %s", e);

                                        debug.error(infoText);
                                        stopThread(infoText);
                                    }
                                } else {
                                    break;
                                }

                                // Nächster Tag:
                                calTagAnfang.add(Calendar.DATE, 1);
                            }
                        } else {
                            stopThread(ABBRUCH);
                        }
                    } else {
                        sendeMeldungInfo("", String.format("Es handelt sich nicht um eine Achslastmessstelle, der Export wird für die Langzeitzählstelle %s abgebrochen", lzz.getPID()));
                    }
                }
            }
        }
    }

    /**
     * Die Methode startet den Exportvorgang für das BASt-Band 2004 sowie den Export der
     * Axlastdaten.
     *
     * @throws Exception .
     */
    private void exportFTPV2004() throws Exception {
        debug.info("Der Export Vorgang V2004 wird gestartet.");

        FTPServerConnection ftpServerConnection = FTPServerConnection.getInstance();

        ftpServerConnection.initialisiereProgressbarKonstanten();

        if (_LzzPidList != null) {
            for (String pid : _LzzPidList) {
                if (!_abbruchExport) {
                    LangZeitZaehlStelle lzz = null;

                    try {
                        lzz = LangZeitZaehlStelle.getInstance(pid);
                    }
                    catch (Exception e) {
                        sendeMeldungFehler("", String.format("Die Langzeitzählstelle mit der Pid %s Konnte nicht ermittelt werden!", pid));
                    }

                    if (lzz == null) {
                        sendeMeldungFehler("", String.format("Es existieren keine Langzeitzählstelle mit der Pid %s. Der Export für den Datensatz findet nicht statt!", pid));
                    } else if (!isAchslastMessstelle(lzz)) {
                        if (!_abbruchExport) {
                            sendeMeldungInfo("", String.format("Die Daten der Langzeitzählstelle mit der Pid: %s werden exportiert.", lzz));

                            // -----------------------------------------------------------------------
                            // Definiere End-Zeitpunkt:
                            Calendar calBis = erstelleKalenderMitMonatAnfang(_CalEnd);

                            // -----------------------------------------------------------------------
                            // Iteriere durch alle Monate:
                            Calendar calMonatAnfang = erstelleKalenderMitMonatAnfang(_CalStart);
                            Calendar calMonatEnde;

                            while (calMonatAnfang.getTimeInMillis() <= calBis.getTimeInMillis()) {

                                //
                                if (!_abbruchExport) {
                                    calMonatEnde = erstelleKalenderMitMonatEnde(calMonatAnfang);

                                    try {
                                        _BastExporter.export(lzz, calMonatAnfang, calMonatEnde);
                                    }
                                    catch (Exception e) {
                                        String infoText = String.format("Es sind Probleme beim Export aufgetreten. Der aktuelle Exportvorgang wird abgebrochen. %s", e);

                                        debug.error(infoText);
                                        stopThread(infoText);
                                    }
                                } else {
                                    stopThread(ABBRUCH);
                                }

                                // Nächster Tag:
                                calMonatAnfang.add(Calendar.MONTH, 1);
                            }
                        } else {
                            stopThread(ABBRUCH);
                        }
                    } else {
                        sendeMeldungInfo("", String.format("Es handelt sich um eine Achslastmessstelle, der Export für die Achslastmessstelle %s wird abgebrochen", lzz.getPID()));
                    }
                } else {
                    stopThread(ABBRUCH);
                }
            }
        } else {
            sendeMeldungInfo("100", "Es existieren keine Langzeitzählstellen. Der Export wird abgebrochen!");
        }
    }

    /**
     * Die Methode starte den export für das BASt-Band Format 2007 auf den FTP Server,
     * die URI wird als Komponente File übergeben.
     *
     * @throws Exception .
     */
    private void exportFTPVV2007() throws Exception {
        debug.info("Der Exportvorgang V2007 wird gestartet.");
        FTPServerConnection.getInstance().initialisiereProgressbarKonstanten();

        if (_LzzPidList != null) {
            for (String pid : _LzzPidList) {
                if (!_abbruchExport) {
                    LangZeitZaehlStelle lzz = null;

                    try {
                        lzz = LangZeitZaehlStelle.getInstance(pid);
                    }
                    catch (Exception e) {
                        sendeMeldungFehler("", String.format("Die Langzeitzählstelle mit der Pid %s Konnte nicht ermittelt werden!", pid));
                    }

                    if (lzz == null) {
                        sendeMeldungFehler("", String.format("Es existieren keine Langzeitzählstelle mit der Pid %s. Der Export für den Datensatz findet nicht statt!", pid));
                    } else if (!isAchslastMessstelle(lzz)) {
                        sendeMeldungInfo("", String.format("Die Daten der Langzeitzählstelle mit der Pid: %s werden exportiert.", lzz));

                        // Export über mehrere Monate ...
                        if (_CalStart.get(Calendar.YEAR) < _CalEnd.get(Calendar.YEAR)) {
                            int startMonth         = _CalStart.get(Calendar.MONTH);
                            int startBisJahresEnde = 11 - startMonth;
                            int endMonth           = _CalEnd.get(Calendar.MONTH);

                            for (int i = 0; i <= startBisJahresEnde; i++) {
                                if (!_abbruchExport) {
                                    _BastExporter.export(lzz, _CalStart.get(Calendar.YEAR), _CalStart.get(Calendar.MONTH) + i);
                                } else {
                                    stopThread(ABBRUCH);
                                }
                            }

                            for (int i = 0; i <= endMonth; i++) {
                                if (!_abbruchExport) {
                                    _BastExporter.export(lzz, _CalEnd.get(Calendar.YEAR), i);
                                } else {
                                    stopThread(ABBRUCH);
                                }
                            }
                        } else {
                            if ((_monatB - _monatV) >= 1) {
                                for (int i = 0; i <= _monatB - _monatV; i++) {
                                    if (!_abbruchExport) {
                                        _BastExporter.export(lzz, _CalStart.get(Calendar.YEAR), _CalStart.get(Calendar.MONTH) + i);
                                    } else {
                                        stopThread(ABBRUCH);
                                    }
                                }
                            } else {
                                _BastExporter.export(lzz, _CalStart.get(Calendar.YEAR), _CalStart.get(Calendar.MONTH));
                            }
                        }
                    } else {
                        sendeMeldungInfo("", String.format("Es handelt sich um eine Achslastmessstelle, der Export für die Achslastmessstelle %s wird abgebrochen", lzz.getPID()));
                    }
                } else {
                    stopThread(ABBRUCH);
                }
            }
        } else {
            if (LangZeitZaehlStelle.getAll().length == 0) {
                sendeMeldungInfo("100", "Exportvorgang beendet da keine Langzeitzählstellen existieren!");
            }
        }
    }

    //~ GET METHODEN ==========================================================

    /**
     * Getter für Eigenschaft 'parKlassifizierungsgruppeBAST'.
     *
     * @return Wert der Eigenschaft 'parKlassifizierungsgruppeBAST'.
     *
     * @see #_parKlassifizierungsgruppeBAST
     */
    public static String getParKlassifizierungsgruppeBAST() {
        return _parKlassifizierungsgruppeBAST;
    }

    public static int getVersionAxlEinzeldaten() {
        return _versionAxlEinzeldaten;
    }

    //~ METHODEN ==============================================================

    @Override
    public void initialize(ClientDavInterface dav) throws Exception {
        _dav = dav;

        // Initialisiere ClientDavInterface:
        NetzClientDavConnection.initDav(dav);

        // Initialisiere alternative Archiv-DaV:
        // siehe Beschreibung zu _davArchivDatenverteiler !
        if (_davArchivDatenverteiler != null) {
            NetzClientDavConnection.initDavArchiv(_davArchivDatenverteiler, _davArchivBenutzer, _davArchivPasswort);

            try {
                NetzClientDavConnection.getInstanceArchiv().getArchive();
            }
            catch (Exception e) {
                debug.error(String.format("Archivdaten können nicht aus DaV<%s><%s><%s> abgefragt werden!",  //
                                          _davArchivDatenverteiler, _davArchivBenutzer, _davArchivPasswort), e);
                System.exit(-1);
            }
        }

        // Initialisiere BAStStatusDaten:
        _bastStatusDaten = new SendeBAStStatusDaten(_typBAStBand, true);
    }

    //~ GET METHODEN ==========================================================

    /**
     * Die Methode prueft ob es sich bei der zu bearbeitenden Langzeitzaehlstelle um
     * eine Achslastmesstelle handelt.
     * @param lzz  Langzeitzaehlstelle, für die geprueft wird, ob es sich um eine
     * Achslastmesstelle handelt.
     * @return true wenn es sich um eine Achslastmesstelle handelt, false sonst.
     */
    private boolean isAchslastMessstelle(LangZeitZaehlStelle lzz) {
        try {
            return lzz.isAchslastErfassung();
        }
        catch (DataModellMismatchException e) {
            String infoText = "Es sind Probleme beim Bestimmen der Langzeitzählstellen bzgl. Achslasterfassung aufgetreten. Der aktuelle Exportvorgang wird abgebrochen!";

            debug.error(infoText);
            stopThread(infoText);
        }

        return false;
    }

    /**
     * Getter für Eigenschaft 'isDateiNameV2004LzzsNr'.
     *
     * @return Wert der Eigenschaft 'isDateiNameV2004LzzsNr'.
     *
     * @see #_isDateiNameV2004LzzsNr
     */
    public static boolean isDateiNameV2004LzzsNr() {
        return _isDateiNameV2004LzzsNr;
    }

    //~ METHODEN ==============================================================

    /**
     * Die Methode konvertiert die Resultdaten in ein internes Format, so dass damit die
     * Methoden für den Export aufgerufen werden können.
     *
     * @param result ResultData vom Datenverteiler mit den Informationen bzgl. des zu startenden
     *               Exportes.
     * @throws DataModellMismatchException .
     */
    private void konvertiereDaVDatenInInternesFormat(ResultData result) throws DataModellMismatchException {
        Data data = result.getData();

        // vonTag - bisTag:
        _tagVon = data.getScaledValue("vonTag").intValue();
        _tagBis = data.getScaledValue("bisTag").intValue();

        // vonMonat
        Data.TextValue vonMonat = data.getTextValue("vonMonat");

        if (data.getTextValue("FTP-Server").getText().equals("EXPORT-ABBRUCH")) {
            _abbruchExport = true;
            stopThread(ABBRUCH);
        } else {
            _abbruchExport = false;

            // vonJahr
            Data.TextValue vonJahr       = data.getTextValue("vonJahr");
            int            monatsWahlVon = 0;
            int            monatsWahlBis = 0;

            _CalStart = Calendar.getInstance();

            String[] ar_monate = { "Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember" };

            try {
                for (int i = 0; i < ar_monate.length; i++) {
                    if (ar_monate[i].equals(vonMonat.getText())) {
                        monatsWahlVon = i;

                        break;
                    }
                }

                SimpleDateFormat dateFormatStart = new SimpleDateFormat("d/M/yyyy/kk/mm");
                Date             dateStart       = Calendar.getInstance().getTime();

                try {
                    _monatV   = monatsWahlVon + 1;
                    dateStart = dateFormatStart.parse("1/" + _monatV + "/" + vonJahr.getText() + "/00/00");
                }
                catch (ParseException e) {
                    debug.error("Es sind Probleme bei der Konvertierung der Daten aufgetreten." + e, e);
                }

                _CalStart.setTime(dateStart);
            }
            catch (NumberFormatException e) {
                debug.error("Es sind Probleme bei der Konvertierung aufgetreten." + e, e);
            }

            // bisMonat
            Data.TextValue bisMonat = data.getTextValue("bisMonat");

            // bisJahr
            Data.TextValue bisJahr = data.getTextValue("bisJahr");

            _CalEnd = Calendar.getInstance();

            try {
                for (int i = 0; i < ar_monate.length; i++) {
                    if (ar_monate[i].equals(bisMonat.getText())) {
                        monatsWahlBis = i;
                    }
                }

                Calendar         cal            = new GregorianCalendar(Integer.parseInt(bisJahr.getText()), monatsWahlBis, 1);
                int              days           = cal.getActualMaximum(Calendar.DAY_OF_MONTH);
                Date             dateEnde       = Calendar.getInstance().getTime();
                SimpleDateFormat dateFormatEnde = new SimpleDateFormat("d/M/yyyy/kk/mm");

                try {
                    _monatB = monatsWahlBis + 1;

                    // Endezeitpunkt
                    // Letze Stunde war 23 und nicht 24 -- dateEnde = dateFormatEnde.parse(days + "/" + _monatB + "/" + bisJahr.getText() + "/23/00");
                    dateEnde = dateFormatEnde.parse(days + "/" + _monatB + "/" + bisJahr.getText() + "/23/59");
                }
                catch (ParseException e) {
                    debug.error("Es sind Probleme bei der Konvertierung des Datums aufgetreten." + e, e);
                }

                _CalEnd.setTime(dateEnde);
            }
            catch (NumberFormatException e) {
                debug.error("Es sind Probleme bei der Konvertierung aufgetreten." + e, e);
            }

            if (_CalStart.get(Calendar.YEAR) < _CalEnd.get(Calendar.YEAR)) {
                int anzahlMonate = (12 - _CalStart.get(Calendar.MONTH)) + 1 + _CalEnd.get(Calendar.MONTH) + 1;

                FTPServerConnection.getInstance().setAnzahlMonate(anzahlMonate);
            } else {
                FTPServerConnection.getInstance().setAnzahlMonate(_monatB - _monatV);
            }

            // Ausgabeformat je nach Wahl im View.
            Data.TextValue bastBandAusgabeFormat = data.getTextValue("BastBandAusgabeFormat");

            _bastFormat = BAST_FORMAT.ermittleFormat(bastBandAusgabeFormat.getText());

            // Ungültiges Format:
            if (_bastFormat == null) {
                sendeMeldungInfo("", "Das übergebene BASt-Band Format existiert nicht, der Export wird abgebrochen.");

                return;
            }

            //
            // DaV-Datenformat: Siehe Methode de.inovat.vew.autoexport.AutoExport.initialisierung()
            // oder de.inovat.buv.plugin.gtm.bast.gui.ExportGUIVerwaltung.sendeExportDaten()
            //
            // FTP-Server Angaben:
            String          ftpServer       = data.getTextValue("FTP-Server").getText();
            ServerParameter serverParameter = new ServerParameter(ftpServer);

            //
            Data.TextValue bastBandUeberschreiben = data.getTextValue("BAStBandUeberschreiben");

            // Angabe ob schon existierende BASt-Band Dateien zur gleichen Selektion überschrieben werden sollen
            String  ueberschreiben = bastBandUeberschreiben.getText();
            boolean dateiErsetzten = false;

            if (ueberschreiben.equals("Ja")) {
                dateiErsetzten = true;
            }

            Data.TextValue aenderungDatenstruktur = data.getTextValue("AenderungDatenstruktur");
            String         aenderung              = aenderungDatenstruktur.getText();
            boolean        aenderungErfolgt       = false;

            if (aenderung.equals("Ja")) {
                aenderungErfolgt = true;
            }

            // LangZeitZählStellen.
            Data.ReferenceArray   refAr_Lzz      = data.getItem("LangZeitZählstellen").asReferenceArray();
            Data.ReferenceValue[] refValueAr_Lzz = refAr_Lzz.getReferenceValues();

            // Füllen der Liste mit allen PID-Nummern der zu exportierenden Langzeitzählstellen.
            _LzzPidList = new ArrayList<>();

            for (Data.ReferenceValue aRefValueAr_Lzz : refValueAr_Lzz) {
                _LzzPidList.add(aRefValueAr_Lzz.getSystemObjectPid());
            }

            FTPServerConnection.getInstance().setAnzahlLZZ(_LzzPidList.size());
            FTPServerConnection.getInstance().setExportiertAnzahl(0);
            FTPServerConnection.getInstance().setParameterExport(serverParameter, _bastFormat, dateiErsetzten, aenderungErfolgt);

            //
            starteThread();
        }
    }

    /**
     * Main.
     *
     * @param args Aufrufparameter.
     */
    public static void main(String[] args) {
        try {

            // ---------------------------------------------
            // Initialisiere das Program:
            BAStExportSteuerung bastExportSteuerung = new BAStExportSteuerung();

            StandardApplicationRunner.run(bastExportSteuerung, args);

            // Debug erst hier initialisieren, da erst ab hier verfügbar !!
            debug = Debug.getLogger();

            // ---------------------------------------------
            // Starte den Export-DaV-Überwachung:
            bastExportSteuerung.sendeMeldungInfo("100", "Der Prozess BASt-Band läuft. Der Export kann gestartet werden.");
            bastExportSteuerung.anmeldenDaVEmpfaengerBAStExport();
            bastExportSteuerung.arbeiteAuftraegeAb();
        }
        catch (Exception e) {
            debug.error("Es ist ein Fehler aufgetreten. Applikation wird beendet.", e);
            System.exit(2);
        }
    }

    @Override
    public void parseArguments(ArgumentList argumentList) throws Exception {
        String par = "-pidBastExportImportSteuerung";

        // --------------------------------------------------------------------
        // Hole den logischen Parameter für die BAStExportImportSteuerung:
        try {
            _typBAStBand = argumentList.fetchArgument(par).asNonEmptyString();
        }
        catch (Exception e) {
            debug.error(String.format("Bitte den Parameter <%s> für die BASt-Export/Import-Steuerung übergeben!", par), e);

            throw new RuntimeException(e);
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -versionAxlEinzeldaten
        par = "-versionAxlEinzeldaten";

        String versionAxlEinzeldatenAufrufParameter = argumentList.fetchArgument(String.format("%s=3", par)).asNonEmptyString().trim();

        try {
            _versionAxlEinzeldaten = new Integer(versionAxlEinzeldatenAufrufParameter);

            if ((_versionAxlEinzeldaten < 0)
                || (_versionAxlEinzeldaten > 3)) {
                throw new NumberFormatException();
            }
        }
        catch (NumberFormatException e) {
            _versionAxlEinzeldaten = 3;
            debug.warning(String.format("Aufrufparameter [%s=%s] ist ungültig. Wert wurde für weiter Berechnung auf [%s=%d] gesetzt.",  //
                                        par, versionAxlEinzeldatenAufrufParameter, par, _versionAxlEinzeldaten));
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter für den DateiNamen V2004:
        par                     = "-dateiNameV2004LzzsNr";
        _isDateiNameV2004LzzsNr = argumentList.fetchArgument(String.format("%s=nein", par)).booleanValue();

        // --------------------------------------------------------------------
        // Hole den logischen Parameter für die BAStExportImportSteuerung:
        par = "-klassifizierungsGruppeBast";

        try {
            _parKlassifizierungsgruppeBAST = argumentList.fetchArgument(par).asNonEmptyString().trim();

            boolean parOk = false;

            switch (_parKlassifizierungsgruppeBAST) {
            case "1" :
            case "2" :
                parOk = true;

                break;

            case "5plus1" :
                _parKlassifizierungsgruppeBAST = "5 plus 1";
                parOk                          = true;

                break;

            case "8plus1" :
                _parKlassifizierungsgruppeBAST = "8 plus 1";
                parOk                          = true;

                break;
            }

            if (!parOk) {
                debug.error(String.format("Der Aufrufparameter ist falsch (%s = %s)! Mögliche Werte: 8plus1, 5plus1, 2, 1.", par, _parKlassifizierungsgruppeBAST));
            } else {
                debug.warning(String.format("Es wird alternative zur DaV-Konfiguration KlassifizierungsgruppeBAST verwendet (%s = %s)!", par, _parKlassifizierungsgruppeBAST));
            }
        }
        catch (Exception e) {
            _parKlassifizierungsgruppeBAST = null;
        }

        // --------------------------------------------------------------------
        // Hole Parameter für die alternative Archiv-DaV:
        // siehe Beschreibung zu _davArchivDatenverteiler !
        // ... Host:
        par = "-davArchivDatenverteiler";

        try {
            _davArchivDatenverteiler = argumentList.fetchArgument(par).asNonEmptyString();
            _davArchivDatenverteiler = String.format("-datenverteiler=%s", _davArchivDatenverteiler);
        }
        catch (Exception e) {
            _davArchivDatenverteiler = null;
        }

        if (_davArchivDatenverteiler != null) {

            // ... Benutzer:
            par = "-davArchivBenutzer";

            try {
                _davArchivBenutzer = argumentList.fetchArgument(par).asNonEmptyString();
            }
            catch (Exception e) {
                debug.error(String.format("Parameter <%s> für den Benutzernamen zur Archiv-DaV ist leer!", par), e);
                _davArchivBenutzer = "";
            }

            // ... Passwort:
            par = "-davArchivPasswort";

            try {
                _davArchivPasswort = argumentList.fetchArgument(par).asNonEmptyString();
            }
            catch (Exception e) {
                debug.error(String.format("Parameter <%s> für das Passwort zur Archiv-DaV ist leer!", par), e);
                _davArchivPasswort = "";
            }
        }
    }

    /**
     * Die Methode sorgt dafür, dass der erzeugte Thread den BastBand Export durchführt.
     */
    public void run() {
        if (!_abbruchExport) {
            sendeMeldungInfo("2", "Der Exportvorgang wird gestartet!");

            if (starteBAStExporter()) {

                // Stoppen um den Thread wieder neu aktivieren zu können.
                stopThread("Der Exportvorgang wurde abgeschlossen!");
            } else {

                // Stoppen um den Thread wieder neu aktivieren zu können.
                stopThread("Der Exportvorgang konnte nicht erfolgreich abgeschlossen werden!");
            }
        } else {
            stopThread(ABBRUCH);
        }
    }

    /**
     * Die Methode sendet die übergebene Fehler_Meldung an den DaV.
     *
     * @param prozent Prozent.
     * @param meldung Fehlermeldung die publiziert wird.
     */
    private void sendeMeldungFehler(String prozent, String meldung) {
        try {
            _bastStatusDaten.sendeBAStStatusDaten(prozent, meldung);
            debug.error(meldung);
        }
        catch (SendSubscriptionNotConfirmed e) {
            debug.error("Es sind Probleme beim Senden der Statusinformationsdaten aufgetreten.", e);
        }
    }

    /**
     * Die Methode sendet die übergebene Info_Meldung an den DaV.
     *
     * @param prozent Prozent.
     * @param meldung Infomeldung die publiziert wird.
     */
    private void sendeMeldungInfo(String prozent, String meldung) {
        try {
            _bastStatusDaten.sendeBAStStatusDaten(prozent, meldung);
            debug.info(meldung);
        }
        catch (SendSubscriptionNotConfirmed e) {
            debug.error("Es sind Probleme beim Senden der Statusinformationsdaten aufgetreten.", e);
        }
    }

    /**
     * Die Methode startet den BASt-Band Export je nach BASt-Band Format.
     *
     * @return boolesche Variable die angibt, ob der Exportvorgang erfolgreich durchgeführt worden ist.
     */
    private boolean starteBAStExporter() {
        boolean erfolgreich = false;

        try {
            _BastExporter = new BastExporter(_bastFormat, _dav, _bastStatusDaten);
            erfolgreich   = true;

            if (BAST_FORMAT.V2007 == _bastFormat) {
                exportFTPVV2007();
            } else if (BAST_FORMAT.V2004 == _bastFormat) {
                exportFTPV2004();
            } else if (BAST_FORMAT.AXL_EF1 == _bastFormat) {
                exportFTPAxlEF1();
            } else {
                erfolgreich = false;
                sendeMeldungFehler("", "Das übergebene BASt-Band Format existiert nicht, der Export wird abgebrochen.");
            }
        }
        catch (Exception e) {
            debug.error("Es sind Probleme bei der Erzeugung der BASt-Instanz aufgetreten.", e);
            sendeMeldungFehler("", "Es sind Probleme beim Export aufgetreten, die zu einem schwerwiegendem Problem geführt haben. Der BASt-Band Prozess wird beendet und muss neu gestartet werden.");
            stopThread("Es sind Probleme bei der Erzeugung der BASt-Instanz aufgetreten." + e);
        }

        return erfolgreich;
    }

    /**
     * Die Methode erzeugt einen Thread, der den BastBandExport startet.
     * Durch die Verwendung eines Threads ist es möglich den Export zu jedem
     * Zeitpunkt zu unterbrechen.
     */
    private void starteThread() {
        if (!_threadAktiv) {
            _abbruchExport        = false;
            _threadBastBandExport = new Thread(this);
            _threadBastBandExport.setName("BASTBandExport");
            _threadAktiv = true;
            _threadBastBandExport.start();
        }
    }

    /**
     * Stoppt den Thread zur Überprüfung des Archivstatus.
     *
     * @param meldung Statusmeldung die an den Clienten via DaV geschickt wird.
     */
    private void stopThread(String meldung) {
        if (_threadAktiv) {

            // noinspection AssignmentToNull
            _threadAktiv = false;
            sendeMeldungInfo("100", meldung);
            _abbruchExport = true;
        }
    }

    /**
     * Anmeldung zum Empfäang der Daten bzgl. BASt-Band Exportsteuerung.
     *
     * @param systemObjectBAStExportImportSteuerung
     *                        SystemObject.
     * @param datenbeschreibungExportSteuerung
     *                        Datenbeschreibung die angemeldet wird.
     * @param anmeldeoptionen Anmeldeoptionen.
     * @param empfaengerrolle Empfängerrolle.
     */
    private void subscribe(SystemObject systemObjectBAStExportImportSteuerung, DataDescription datenbeschreibungExportSteuerung, ReceiveOptions anmeldeoptionen, ReceiverRole empfaengerrolle) {
        if (systemObjectBAStExportImportSteuerung == null) {

            // Fehlermeldung via Statusmeldung
        } else {
            try {
                _dav.subscribeReceiver(this, systemObjectBAStExportImportSteuerung, datenbeschreibungExportSteuerung, anmeldeoptionen, empfaengerrolle);
            }
            catch (Exception e) {
                String infoText = "Es sind Probleme bei der Anmeldung beim Datenverteiler aufgetreten. Die Anwendung wird geschlossen." + e;

                debug.error(infoText);
                stopThread(infoText);
            }
        }
    }

    /**
     * Update-Methode wird vom Datenverteiler aufgerufen. Die übergebenen Datensätze werden
     * verarbeitet.
     *
     * @param resultDatas Resultdata, entspricht den Attributen eines BASt-Band exportes,
     *                    d.h. Referenz auf Langzeitzählstellen
     *                    username
     *                    passwort
     *                    url zum FTP-Server auf dem die Exportierten Daten abgelegt werden.
     */
    public void update(ResultData[] resultDatas) {
        for (ResultData result : resultDatas) {
            if (result.hasData()) {
                synchronized (_listeAuftraege) {
                    _listeAuftraege.add(result);
                }
            }
        }
    }

    //~ INNERE KLASSEN ========================================================

    /**
     * Parameter für den Server.
     */
    static class ServerParameter {

        /** Server Benutzername. */
        private String _benutzer = "";

        /** Server Passwort. */
        private String _passwd = "";

        /** FTP-Servername. */
        private String _host = "";

        /** Serverport. */
        private int _port = 0;

        /** Definiert Datei (private Key) für die SFTP-Verbindung bei der Schlüssel-basierter Authentifizierung. */
        private String _prvKeyDatei = "";

        /** Definiert Passwort für private Key. */
        private String _prvKeyPasswort = "";

        /** Definiert die Speicherart der Dateien auf dem Server  (true -hierarchisch, false - flach). */
        private boolean _speicherungHierarchisch = true;

        /** Legt fest, ob die FTP-Verbindung im PassivModus eingebaut werden soll. */
        private boolean _isPassivmodus;

        /** Definiert den ServerTyp: FTP oder SFTP: falls true - FTP, sonst - SFTP. */
        private boolean _istFtp;

        //~ KONSTRUKTOREN  (und vom Konstruktor verwendete Methoden) ==========

        /**
         * Erstellt eine Instanz der Klasse.
         *
         * @param txtServer Serverparameter in der Text-Form (im DaV-Text-Attribut 'FTP-Server').
         */
        ServerParameter(String txtServer) {

            // Server Angaben:
            String passivmodus = "";
            String servertyp   = "";

            // 3.Variante (SFTP mit der Schlüssel-basierter Authentifizierung; Speicherung hierarchisch/flach)
            Pattern pattern = Pattern.compile("-Benutzer:(.*)-Passwort:(.*)-URL:(.*)-Port:(.*)-Passivmodus:(.*)-Servertyp:(.*)-PrvKeyDatei:(.*)-PrvKeyPasswort:(.*)-SpeicherungHierarchisch:(.*)");
            Matcher matcher = pattern.matcher(txtServer);

            if (matcher.find()) {
                _benutzer = matcher.group(1);
                _passwd   = matcher.group(2);
                _host     = matcher.group(3);

                try {
                    _port = Integer.parseInt(matcher.group(4));
                }
                catch (Exception e) {
                    debug.error("Es sind Probleme bem Auslesen des FTP-Server-Ports aus dem DaV-Attribut aufgetreten.");
                }

                passivmodus     = matcher.group(5);
                servertyp       = matcher.group(6);
                _prvKeyDatei    = matcher.group(7);
                _prvKeyPasswort = matcher.group(8);

                try {
                    String speicherungHierarchisch = matcher.group(9);

                    _speicherungHierarchisch = !speicherungHierarchisch.toUpperCase().equals("NEIN");
                }
                catch (Exception e) {
                    debug.error("Es sind Probleme bem Auslesen der Speicherart aus dem DaV-Attribut aufgetreten.");
                }
            }

            // 2.Variante (+ SFTP)
            else {
                pattern = Pattern.compile("-Benutzer:(.*)-Passwort:(.*)-URL:(.*)-Port:(.*)-Passivmodus:(.*)-Servertyp:(.*)");
                matcher = pattern.matcher(txtServer);

                if (matcher.find()) {
                    _benutzer = matcher.group(1);
                    _passwd   = matcher.group(2);
                    _host     = matcher.group(3);

                    try {
                        _port = Integer.parseInt(matcher.group(4));
                    }
                    catch (Exception e) {
                        debug.error("Es sind Probleme bem Auslesen des FTP-Server-Port aus dem DaV-Attribut aufgetreten.");
                    }

                    passivmodus = matcher.group(5);
                    servertyp   = matcher.group(6);
                }

                // 1. Variante (alt):
                else {
                    pattern = Pattern.compile("Benutzer:(.*)Passwort:(.*)URL:(.*)Passivmodus:(.*)");
                    matcher = pattern.matcher(txtServer);

                    if (matcher.find()) {
                        _benutzer   = matcher.group(1);
                        _passwd     = matcher.group(2);
                        _host       = matcher.group(3);
                        _port       = 21;
                        passivmodus = matcher.group(4);
                        servertyp   = "FTP";
                    } else {
                        debug.error("Es sind Probleme bem Auslesen der FTP-Server-Einstellungen aus dem DaV-Attribut aufgetreten.");
                    }
                }
            }

            //
            _isPassivmodus = (passivmodus.toLowerCase().equals("ja"));
            _istFtp        = servertyp.toUpperCase().equals("FTP");
        }

        //~ GET METHODEN ======================================================

        /**
         * Getter für Eigenschaft 'benutzer'.
         *
         * @return Wert der Eigenschaft 'benutzer'.
         *
         * @see #_benutzer
         */
        public String getBenutzer() {
            return _benutzer;
        }

        /**
         * Getter für Eigenschaft 'host'.
         *
         * @return Wert der Eigenschaft 'host'.
         *
         * @see #_host
         */
        public String getHost() {
            return _host;
        }

        /**
         * Getter für Eigenschaft 'passwd'.
         *
         * @return Wert der Eigenschaft 'passwd'.
         *
         * @see #_passwd
         */
        public String getPasswd() {
            return _passwd;
        }

        /**
         * Getter für Eigenschaft 'port'.
         *
         * @return Wert der Eigenschaft 'port'.
         *
         * @see #_port
         */
        public int getPort() {
            return _port;
        }

        /**
         * Getter für Eigenschaft 'prvKeyDatei'.
         *
         * @return Wert der Eigenschaft 'prvKeyDatei'.
         *
         * @see #_prvKeyDatei
         */
        public String getPrvKeyDatei() {
            return _prvKeyDatei;
        }

        /**
         * Getter für Eigenschaft 'prvKeyPasswort'.
         *
         * @return Wert der Eigenschaft 'prvKeyPasswort'.
         *
         * @see #_prvKeyPasswort
         */
        public String getPrvKeyPasswort() {
            return _prvKeyPasswort;
        }

        /**
         * Getter für Eigenschaft 'istFtp'.
         *
         * @return Wert der Eigenschaft 'istFtp'.
         *
         * @see #_istFtp
         */
        public boolean isIstFtp() {
            return _istFtp;
        }

        /**
         * Getter für Eigenschaft 'isPassivmodus'.
         *
         * @return Wert der Eigenschaft 'isPassivmodus'.
         *
         * @see #_isPassivmodus
         */
        public boolean isPassivmodus() {
            return _isPassivmodus;
        }

        /**
         * Getter für Eigenschaft 'speicherungHierarchisch'.
         *
         * @return Wert der Eigenschaft 'speicherungHierarchisch'.
         *
         * @see #_speicherungHierarchisch
         */
        public boolean isSpeicherungHierarchisch() {
            return _speicherungHierarchisch;
        }
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 02.07.19
