/*
 * Copyright (c) 2010-2012 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.inovat.kex.bast.importer.SFTPServerKommunikation
 *
 * de.inovat.kex.bast.importer.SFTPServerKommunikation is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * de.inovat.kex.bast.importer.SFTPServerKommunikation is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * <p>
 * You should have received a copy of the GNU General Public License
 * along with de.inovat.kex.bast.importer.SFTPServerKommunikation; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * Koelner Strasse 30
 * D-50859 Koeln
 * +49 (0)2234 4301 800
 * info@invat.de
 * www.inovat.de
 */



package de.inovat.kex.bast.importer;

//~ NICHT JDK IMPORTE =========================================================

import com.jcraft.jsch.Session;
import de.inovat.kex.bast.serverfunc.SFTPFunktionen;

import java.io.*;
import java.util.List;

//~ JDK IMPORTE ===============================================================

//~ KLASSEN ===================================================================

/**
 * Klasse realisiert die SFTP-Server-Kommunikation beim BASt-Band-Import
 * (s. {@link de.inovat.kex.bast.importer.BAStImportSteuerung#run()}).
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Liliya Givorgizova (LG)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class SFTPServerKommunikation implements IServerKommunikation {

    /**
     * WICHTIG:
     * Die SFTP-Datei wird in die lokale Datei in den Temp-Ordner übertragen (s.{@link #ermittleDateiInputStream(String)}),
     * in {@link BAStImportSteuerung#konvertiereDateiInBastFormat(String, java.io.InputStream)} bearbeitet
     * und in {@link #sperreDateiAufheben(String)} gelöscht.
     */
    private String _dateiLokal;

    /** SFTP Session. */
    private Session _session;

    /** SFTP-Benutzername. */
    private String _sftpBenutzer;

    /** SFTP-Ordner, wo die zu bearbeitenden Dateien liegen. */
    private String _sftpOrdner;

    /** SFTP-Passwort. */
    private String _sftpPasswd;

    /** Host des SFTP-Servers. */
    private String _sftpServerName;

    /** Port des SFTP-Servers. */
    private int _sftpServerPort;

    /** Lokaler Temp-Ordner. */
    private String _tempOrdner;

    //~ METHODEN ==============================================================

    @Override
    public InputStream ermittleDateiInputStream(String dateiPfad) {

        // Übertrage die SFTP-Datei in die lokale Datei und übergebe diese als FileInputStream :
        try {
            if (_sftpOrdner.isEmpty()) {
                _dateiLokal = String.format("%s.tmp", dateiPfad);
            } else {
                _dateiLokal = String.format("%s.tmp",  //
                                            dateiPfad.substring(dateiPfad.indexOf(_sftpOrdner) + _sftpOrdner.length() + 1));
            }

            _dateiLokal = String.format("%s%s", _tempOrdner, _dateiLokal);

            Exception exception = SFTPFunktionen.uebertrageSFTPDateiInLokaleDatei(_session, dateiPfad, _dateiLokal);

            if (exception != null) {
                return null;
            }

            return new FileInputStream(_dateiLokal);
        }
        catch (Exception e) {
            return null;
        }
    }

    /**
     * Lösche die temporale lokale Datei (s. {@link #_dateiLokal}).
     */
    private void loescheLokaleDatei() {
        try {
            new File(_dateiLokal).delete();
        }
        catch (Exception ignored) {}
    }

    @Override
    public void parameterUebergabe(String tempOrdner, String host, int port, String benutzer, String passwort, boolean passivModus, String ordner) {
        _tempOrdner     = tempOrdner;
        _sftpServerName = host;
        _sftpServerPort = port;
        _sftpBenutzer   = benutzer;
        _sftpPasswd     = passwort;
        _sftpOrdner     = ordner;
    }

    @Override
    public Exception sperreDatei(String dateiMitPfad) {

        // Trenne alte Verbindung und verbinde erneut (sonst nach ca. 3._Fehler Exception 'Too many users...'):
        Exception exception = verbindeMitServer();

        if (exception != null) {
            return exception;
        }

        // Ermittle alle Dateien im SFTP-Ordner:
        List<String> listeDateien = SFTPFunktionen.ermittleDateienImSFTPOrdner(_session, _sftpOrdner);

        // Suche die angegebene Datei:
        if (listeDateien == null) {
            return new Exception("Die SFTP-Verzeichnisinformationsdaten können nicht angezeigt werden.");
        } else if (listeDateien.isEmpty()) {
            return new Exception("Es sind keine Dateien im SFTP-Verzeichnis vorhanden.");
        } else {
            try {
                String dateiSftp;

                for (String name : listeDateien) {

                    // Ergänze die SFTP-Datei mit Ordner:
                    dateiSftp = String.format("%s%s%s", _sftpOrdner, BAStImportSteuerung.FILESEPARATOR_SERVER, name);

                    // Sperre die Datei:
                    if (dateiSftp.equals(dateiMitPfad)) {
                        return SFTPFunktionen.verschiebeSFTPDatei(_session, dateiMitPfad, String.format("%s%s", dateiMitPfad, BAStImportSteuerung.DATEI_LOCK));
                    }

                    // Lock-Datei vorhanden:
                    else if (dateiSftp.equals(String.format("%s%s", dateiMitPfad, BAStImportSteuerung.DATEI_LOCK))) {
                        return new Exception(String.format(  //
                            "Die Datei <%s> konnte nicht gesperrt werden, da die Datei schon gesperrt ist!", dateiMitPfad));
                    }

                    // Bearbeitungs-Datei vorhanden:
                    else if (dateiSftp.equals(String.format("%s%s", dateiMitPfad, BAStImportSteuerung.DATEI_BEARBEITET))) {
                        return new Exception(String.format("Die Datei <%s> ist schon importiert worden!", dateiMitPfad));
                    }
                }
            }
            catch (Exception e) {
                return new Exception("Die FTP-Verzeichnisinformationsdaten können nicht angezeigt werden.", e);
            }

            return new Exception(String.format(              //
                "Die angegebene Datei <%s> befindet sich nicht in dem Importverzeichnis! Bitte überprüfen Sie ihre Angaben!", dateiMitPfad));
        }
    }

    @Override
    public Exception sperreDateiAufheben(String dateiMitPfad) {

        // -----------------------------------------------
        // Lösche die temporale lokale Datei:
        loescheLokaleDatei();

        // -----------------------------------------------
        // Trenne alte Verbindung und verbinde erneut (sonst nach ca. 3._Fehler Exception 'Too many users...'):
        Exception exception = verbindeMitServer();

        if (exception != null) {
            return exception;
        }

        // Umbenennen der Datei:
        try {
            String dateiAlt = String.format("%s%s", dateiMitPfad, BAStImportSteuerung.DATEI_LOCK);
            String dateiNeu = String.format("%s%s", dateiMitPfad, BAStImportSteuerung.DATEI_BEARBEITET);

            return SFTPFunktionen.verschiebeSFTPDatei(_session, dateiAlt, dateiNeu);
        }
        catch (Exception e) {
            return new Exception(String.format("Die Sperre der Datei <%s> kann nicht aufgehoben werden.", dateiMitPfad), e);
        }
    }

    @Override
    public Exception trenneVerbindungMitServer() {
        try {
            _session.disconnect();

            return null;
        }
        catch (Exception e) {
            return e;
        }
    }

    @Override
    public Exception verbindeMitServer() {

        // -----------------------------------------
        // Disconnnect:
        try {
            _session.disconnect();
        }
        catch (Exception ignored) {}

        // -----------------------------------------
        // Verbindung:
        try {
            _session = SFTPFunktionen.verbindeMitSFTPServer(_sftpServerName, _sftpServerPort, _sftpBenutzer, _sftpPasswd);
        }
        catch (Exception e) {

            // _Fehler:
            return e;
        }

        // OK :
        return null;
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 27.06.13
